from typing import List, Optional

from oold.backend.interface import SetResolverParam, set_resolver
from oold.backend.sparql import LocalSparqlResolver


def _run(pydantic_version):
    if pydantic_version == "v1":
        from pydantic.v1 import Field

        # based on pydantic v1
        from oold.model.v1 import LinkedBaseModel

        class Entity(LinkedBaseModel):
            class Config:
                schema_extra = {
                    "@context": {
                        # aliases
                        "id": "@id",
                        "type": "@type",
                        # prefixes
                        "schema": "https://schema.org/",
                        "ex": "https://example.com/",
                        # literal property
                        "name": "schema:name",
                    },
                    "iri": "Entity.json",  # the IRI of the schema
                }

            type: Optional[str] = "ex:Entity.json"
            name: str

            def get_iri(self):
                return "ex:" + self.name

        class Person(Entity):
            class Config:
                schema_extra = {
                    "@context": [
                        "Entity.json",  # import the context of the parent class
                        {
                            # object property definition
                            "knows": {
                                "@id": "schema:knows",
                                "@type": "@id",
                                "@container": "@set",
                            }
                        },
                    ],
                    "iri": "Person.json",
                }

            type: Optional[str] = "ex:Person.json"
            knows: Optional[List["Person"]] = Field(
                None,
                # object property pointing to another Person
                range="Person.json",
            )

    if pydantic_version == "v2":
        from pydantic import ConfigDict, Field

        # based on pydantic v2
        from oold.model import LinkedBaseModel  # noqa

        class Entity(LinkedBaseModel):
            model_config = ConfigDict(
                json_schema_extra={
                    "@context": {
                        # aliases
                        "id": "@id",
                        "type": "@type",
                        # prefixes
                        "schema": "https://schema.org/",
                        "ex": "https://example.com/",
                        # literal property
                        "name": "schema:name",
                    },
                    "iri": "Entity.json",  # the IRI of the schema
                }
            )
            type: Optional[str] = "ex:Entity.json"
            name: str

            def get_iri(self):
                return "ex:" + self.name

        class Person(Entity):  # noqa
            model_config = ConfigDict(
                json_schema_extra={
                    "@context": [
                        "Entity.json",  # import the context of the parent class
                        {
                            # object property definition
                            "knows": {
                                "@id": "schema:knows",
                                "@type": "@id",
                                "@container": "@set",
                            }
                        },
                    ],
                    "iri": "Person.json",
                }
            )
            type: Optional[str] = "ex:Person.json"
            knows: Optional[List["Person"]] = Field(
                None,
                # object property pointing to another Person
                json_schema_extra={"range": "Person.json"},
            )

    p1 = Person(name="Alice")
    p2 = Person(name="Bob", knows=[p1])
    print(p2.to_jsonld())

    # load the rdf into a rdflib graph
    from rdflib import Graph

    g = Graph()
    g.parse(data=p1.to_jsonld(), format="json-ld")
    g.parse(data=p2.to_jsonld(), format="json-ld")
    print(g.serialize(format="turtle"))

    # query the name of persons that Bob knows
    qres = g.query(
        """
        SELECT ?name
        WHERE {
            ?s <https://schema.org/knows> ?o .
            ?o <https://schema.org/name> ?name .
        }
        """
    )
    for row in qres:
        print("Bob knows " + row.name)
        assert str(row.name) == "Alice"

    # create a resolver to resolve IRIs to objects
    r = LocalSparqlResolver(graph=g)
    set_resolver(SetResolverParam(iri="ex", resolver=r))

    # load bob from the graph
    bob = Person["ex:Bob"]
    print(bob)
    # accessing 'knows' will trigger a sparql query
    # to get the full node of Alice from the graph
    print(bob.knows[0].name)
    assert bob.knows[0].name == "Alice"


def test_rdf_export_and_sparql_query():
    _run("v1")
    _run("v2")


if __name__ == "__main__":
    test_rdf_export_and_sparql_query()



"""Utilities for working different enviroments like IPython/Jupyter notebooks."""

import ast
import inspect
import sys
from typing import Callable


def is_running_in_ipython() -> bool:
    """Check if the code is running in an interactive
    environment like Jupyter or IPython."""
    return "IPython" in sys.modules


def is_running_in_notebook() -> bool:
    """Check if the code is running in a Jupyter notebook."""
    return "ipykernel" in sys.modules


def is_running_in_pyodide() -> bool:
    """Check if the code is running in a Pyodide environment."""
    return sys.platform == "emscripten"


def get_ipython_source() -> str:
    """Get the source code of the running notebook."""
    from IPython import get_ipython

    shell = get_ipython()
    if shell is None:
        raise RuntimeError("Not running in an IPython environment")

    if not hasattr(shell, "user_ns"):
        raise AttributeError("Cannot access user namespace")

    # list of input cells in the notebook
    input_list = shell.user_ns["In"]

    source_text = "\n\n".join(cell for cell in input_list[1:] if cell)

    return source_text


def get_object_source_from_ipython_source(obj: Callable) -> str:
    """Get the most recent definition of an object in a ipython session."""
    notebook_source = get_ipython_source()
    tree = ast.parse(notebook_source)

    obj_name = obj.__name__

    # Walk the entire tree and get the last matching definition
    segment = None
    for node in ast.walk(tree):
        # Check for function or class definitions, depending on the type of obj
        if (
            isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef))
            and node.name == obj_name
        ):
            segment = ast.get_source_segment(notebook_source, node)
        elif isinstance(node, ast.ClassDef) and node.name == obj_name:
            segment = ast.get_source_segment(notebook_source, node)

    if segment is not None:
        return segment

    raise ValueError(
        f"Object '{obj_name}' definition not found in the ipython session source code"
    )


def get_object_source(obj: Callable) -> str:
    """Get the source code of an object, works in scripts and ipython sessions."""
    try:
        source = inspect.getsource(obj)
    except OSError:
        # no source code file available
        try:
            if is_running_in_ipython():
                source = get_object_source_from_ipython_source(obj)
            else:
                return None
        except ValueError:
            # no class source available
            return None
    return source



# generated by datamodel-codegen:
#   filename:  example.json

from __future__ import annotations

from enum import Enum

from pydantic import ConfigDict, Field

from oold.model import LinkedBaseModel


class Hobby(str, Enum):
    """
    Defines various hobbies as an enum.
    """

    SPORTS = "ex:sports"
    """
    Sports hobby, e.g. football, basketball, etc.
    """
    MUSIC = "ex:music"
    """
    Music hobby, e.g. playing instruments, singing, etc.
    """
    ART = "ex:art"
    """
    Art hobby, e.g. painting, drawing, etc.
    """


class Example(LinkedBaseModel):
    model_config = ConfigDict(
        json_schema_extra={"title": "Example"},
    )
    type: str | None = ["example"]
    hobby: Hobby | None = Field(
        None,
        json_schema_extra={
            "x-enum-varnames": ["SPORTS", "MUSIC", "ART"],
            "options": {
                "enum_titles": [
                    "Sports hobby, e.g. football, basketball, etc.",
                    "Music hobby, e.g. playing instruments, singing, etc.",
                    "Art hobby, e.g. painting, drawing, etc.",
                ]
            },
        },
    )
    """
    Defines various hobbies as an enum.
    """



from static import _run

from oold.static import enum_docstrings as parse_enum_docstrings


def test_oneof_subschema():
    # json schema with property that contains a oneOf with two subschemas

    schemas = [
        {
            "id": "example",
            "title": "Example",
            "type": "object",
            "properties": {
                "type": {"type": "string", "default": ["example"]},
                "prop1": {"type": "string", "custom_key": "custom_value"},
                "prop2": {
                    "custom_key": "custom_value",
                    "properties": {
                        "subprop0": {"type": "string", "custom_key": "custom_value_0"},
                    },
                    "oneOf": [
                        {
                            "title": "Subschema1",
                            "type": "object",
                            "properties": {
                                "subprop1": {
                                    "type": "string",
                                    "custom_key": "custom_value_1",
                                },
                            },
                        },
                        {
                            "title": "Subschema2",
                            "type": "object",
                            "properties": {
                                "subprop2": {
                                    "type": "string",
                                    "custom_key": "custom_value_2",
                                },
                            },
                        },
                    ],
                },
            },
        },
    ]

    def oneof_subschema(pydantic_version):
        # Test the generated model, see
        # https://github.com/koxudaxi/datamodel-code-generator/issues/2403

        if pydantic_version == "v1":
            import data.oneof_subschema.model_v1 as model

            assert (
                model.Subschema1.__fields__["subprop1"].field_info.extra["custom_key"]
                == "custom_value_1"
            )
        else:
            import data.oneof_subschema.model_v2 as model

            model.Subschema1.model_fields["subprop1"].json_schema_extra[
                "custom_key"
            ] == "custom_value_1"

    _run(
        schemas,
        main_schema="example.json",
        test=oneof_subschema,
        # pydantic_versions=["v1"],
    )


def test_enum_docstrings():
    schemas = [
        {
            "id": "example",
            "title": "Example",
            "type": "object",
            "properties": {
                "type": {"type": "string", "default": ["example"]},
                "hobby": {
                    "type": "string",
                    "enum": ["ex:sports", "ex:music", "ex:art"],
                    "description": "Defines various hobbies as an enum.",
                    "x-enum-varnames": ["SPORTS", "MUSIC", "ART"],
                    "options": {
                        "enum_titles": [
                            "Sports hobby, e.g. football, basketball, etc.",
                            "Music hobby, e.g. playing instruments, singing, etc.",
                            "Art hobby, e.g. painting, drawing, etc.",
                        ]
                    },
                },
            },
        },
    ]

    def enum_docstrings(pydantic_version):
        # Test the generated model, see
        # https://github.com/koxudaxi/datamodel-code-generator/issues/2403

        if pydantic_version == "v1":
            import data.enum_docstrings.model_v1 as model

            Hobby = parse_enum_docstrings(model.Hobby)
            assert (
                Hobby.SPORTS.__doc__.strip()
                == "Sports hobby, e.g. football, basketball, etc."
            )
            assert (
                Hobby.MUSIC.__doc__.strip()
                == "Music hobby, e.g. playing instruments, singing, etc."
            )
            assert (
                Hobby.ART.__doc__.strip() == "Art hobby, e.g. painting, drawing, etc."
            )
        else:
            import data.enum_docstrings.model_v2 as model

            Hobby = parse_enum_docstrings(model.Hobby)
            assert (
                Hobby.SPORTS.__doc__.strip()
                == "Sports hobby, e.g. football, basketball, etc."
            )
            assert (
                Hobby.MUSIC.__doc__.strip()
                == "Music hobby, e.g. playing instruments, singing, etc."
            )
            assert (
                Hobby.ART.__doc__.strip() == "Art hobby, e.g. painting, drawing, etc."
            )

    _run(
        schemas,
        main_schema="example.json",
        test=enum_docstrings,
        # pydantic_versions=["v1"],
    )


def test_subclass_inheritance():
    """
    Test that subclass inheritance of properties from parent classes works correctly.
    """

    schemas = [
        {
            "id": "Thing",
            "title": "Thing",
            "type": "object",
            "required": ["type", "name"],
            "properties": {
                "type": {
                    "type": "string",
                    "default": "playground:Thing",
                    "options": {"hidden": True},
                },
                "name": {
                    "type": "string",
                    "description": "The things name",
                    "minLength": 1,
                    "default": "A Thing",
                },
            },
        },
        {
            "id": "Person",
            "title": "Person",
            "type": "object",
            "allOf": [{"$ref": "Thing.json"}],
            "required": ["name"],
            "properties": {
                "type": {
                    "$comment": "Already defined in playground:Thing -> we override just the default",  # noqa: E501
                    "default": "playground:Person",
                },
                "name": {
                    "description": "First and Last name",
                    "minLength": 4,
                    "default": "John Doe",
                },
                "age": {"type": "integer"},
            },
        },
    ]

    def subclass_inheritance(pydantic_version):
        # 'Person' should be a subclass of 'Thing' and inherit its properties
        # the default value of 'name' in 'Thing' should be 'A Thing'
        # the default value of 'name' in 'Person' should be 'John Doe'
        # the type of 'name' in 'Person' should be string

        if pydantic_version == "v1":
            import data.subclass_inheritance.model_v1 as model

            assert issubclass(model.Person, model.Thing)
            assert model.Thing.__fields__["name"].default == "A Thing"
            assert model.Person.__fields__["name"].default == "John Doe"
            # fails with datamodel-code-generator 0.28.2, fixed in 0.43.1
            assert str(model.Person.__fields__["type"].annotation) == "str | None"

        else:
            import data.subclass_inheritance.model_v2 as model

            assert issubclass(model.Person, model.Thing)
            assert model.Thing.model_fields["name"].default == "A Thing"
            assert model.Person.model_fields["name"].default == "John Doe"
            assert str(model.Person.model_fields["type"].annotation) == "str | None"

    _run(
        schemas,
        main_schema="Person.json",
        test=subclass_inheritance,
    )


def test_class_hierarchy():
    schemas = [
        {
            "id": "NestedSubSchema",
            "title": "NestedSubSchema",
            "description": "An example subschema",
            "type": "object",
            "properties": {
                "subschema": {
                    "title": "SimpleSubSchema",
                    "type": "object",
                    "properties": {"some_property": {"type": "string"}},
                }
            },
        },
        {
            "id": "OtherEntity",
            "type": "object",
            "properties": {
                "nested": {
                    "type": "array",
                    "items": {
                        "title": "NestedSubSchema",
                        "x-custom-annotation": "custom value",
                        # "allOf": [{
                        "$ref": "NestedSubSchema.json"
                        # }]
                    },
                }
            },
        },
        {
            "id": "Entity",
            "title": "Entity",
            "type": "object",
            "properties": {
                "nested": {
                    "title": "EntityNestedSubSchema",
                    "description": "An example using a subschema",
                    "x-custom-annotation": "custom value",
                    # "allOf": [{
                    "$ref": "NestedSubSchema.json"
                    # }]
                }
            },
        },
    ]

    def class_hierarchy(pydantic_version):
        # test if the generated modul contains exactly three classes:
        # Entity, NestedSubSchema, SimpleSubSchema

        if pydantic_version == "v1":
            import data.class_hierarchy.model_v1 as model

        else:
            import data.class_hierarchy.model_v2 as model

        assert hasattr(model, "Entity")
        assert hasattr(model, "NestedSubSchema")
        assert hasattr(model, "SimpleSubSchema")
        # ignore imported classes and only count classes defined in the module
        classes = [
            cls_name
            for cls_name in dir(model)
            if isinstance(getattr(model, cls_name), type)
            and getattr(model, cls_name).__module__ == model.__name__
        ]
        assert set(classes) == {"Entity", "NestedSubSchema", "SimpleSubSchema"}

        # test if type annotation of Entity.nested is Optional[NestedSubSchema]
        # and NestedSubSchema.subschema is Optional[SimpleSubSchema]
        if pydantic_version == "v1":
            import data.class_hierarchy.model_v1 as model

            assert (
                str(model.Entity.__fields__["nested"].annotation)
                == "data.class_hierarchy.model_v1.NestedSubSchema | None"
            )
            assert (
                str(model.NestedSubSchema.__fields__["subschema"].annotation)
                == "data.class_hierarchy.model_v1.SimpleSubSchema | None"
            )
        else:
            import data.class_hierarchy.model_v2 as model

            assert (
                str(model.Entity.model_fields["nested"].annotation)
                == "data.class_hierarchy.model_v2.NestedSubSchema | None"
            )
            assert (
                str(model.NestedSubSchema.model_fields["subschema"].annotation)
                == "data.class_hierarchy.model_v2.SimpleSubSchema | None"
            )

    _run(
        schemas,
        main_schema="Entity.json",
        test=class_hierarchy,
    )


if __name__ == "__main__":
    test_oneof_subschema()
    test_enum_docstrings()
    test_subclass_inheritance()
    test_class_hierarchy()



# generated by datamodel-codegen:
#   filename:  Entity.json

from __future__ import annotations

from pydantic.v1 import Field

from oold.model.v1 import LinkedBaseModel


class SimpleSubSchema(LinkedBaseModel):
    class Config:
        schema_extra = {"title": "SimpleSubSchema"}

    some_property: str | None = None


class NestedSubSchema(LinkedBaseModel):
    """
    An example subschema
    """

    class Config:
        schema_extra = {
            "title": "NestedSubSchema",
            "description": "An example subschema",
        }

    subschema: SimpleSubSchema | None = Field(None, title="SimpleSubSchema")


class Entity(LinkedBaseModel):
    class Config:
        schema_extra = {"title": "Entity"}

    nested: NestedSubSchema | None = Field(
        None, title="EntityNestedSubSchema", x_custom_annotation="custom value"
    )
    """
    An example using a subschema
    """



# generated by datamodel-codegen:
#   filename:  Foo.json

from __future__ import annotations

from enum import Enum
from typing import Optional

from pydantic import ConfigDict, Field

from oold.model import LinkedBaseModel


class LengthUnit(str, Enum):
    meter = "Item:OSWf101d25e944856e3bd4b4c9863db7de2"
    milli_meter = (
        "Item:OSWf101d25e944856e3bd4b4c9863db7de2#OSW322dec469be75aedb008b3ebff29db86"
    )
    kilo_meter = (
        "Item:OSWf101d25e944856e3bd4b4c9863db7de2#OSWb1de8f91f1275572b37c2edfe40d5de6"
    )


class Bar2(LinkedBaseModel):
    model_config = ConfigDict(
        json_schema_extra={
            "@context": {
                "id": "@id",
                "type": "@type",
                "ex": "http://example.com/",
                "@base": "http://example.com/",
                "prop1": "ex:prop1",
            },
            "uuid": "bar2/Bar2.json",
            "title": "Bar2",
        },
    )
    id: Optional[str] = None
    type: Optional[list[str]] = ["Bar2"]
    prop1: Optional[str] = None
    unit: Optional[LengthUnit] = Field(
        None,
        json_schema_extra={
            "defaut": "Item:OSWf101d25e944856e3bd4b4c9863db7de2",
            "$comment": "enum_titles are valid variable names and follow latex SI unit notation with '\\' replaced with '_'",
            "x-enum-varnames": ["meter", "milli_meter", "kilo_meter"],
            "enum_titles*": {
                "$comment": "Human friedly symbols",
                "en": ["m", "mm", "km"],
            },
        },
        title="LengthUnit",
    )


class Bar(Bar2):
    model_config = ConfigDict(
        json_schema_extra={"@context": ["./bar2/Bar2.json"], "title": "Bar"},
    )
    type: Optional[list[str]] = ["Bar"]
    prop2: Optional[str] = None
    unit: Optional[str] = Field(
        None,
        json_schema_extra={"_allOf": [{"$ref": "./bar2/Bar2.json#/properties/unit"}]},
        title="DiameterUnit",
    )


class Foo(LinkedBaseModel):
    model_config = ConfigDict(
        json_schema_extra={"title": "Foo"},
    )
    id: Optional[str] = None
    type: Optional[list[str]] = ["Foo"]
    literal: Optional[str] = None
    b: Optional[Bar] = Field(None, json_schema_extra={"range": "Bar.json"})
    b2: Optional[list[Bar]] = Field(None, json_schema_extra={"range": "Bar.json"})



import json
from typing import Dict, List, Optional

from pydantic import ConfigDict
from rdflib import Graph
from SPARQLWrapper import JSONLD, SPARQLWrapper

from oold.backend.auth import GetCredentialParam, UserPwdCredential, get_credential
from oold.backend.interface import Backend, Resolver, StoreResult


class LocalSparqlResolver(Resolver):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    graph: Optional[Graph] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.graph is None:
            self.graph = Graph()

    def resolve_iris(self, iris: List[str]) -> Dict[str, Dict]:
        # sparql query to get a node by IRI with all its properties
        # using CONSTRUCT to get the full node
        # format the result as json-ld
        jsonld_dicts = {}
        for iri in iris:
            iri_filter = f"FILTER (?s = {iri})"
            # check if the iri is a full IRI or a prefix
            if iri.startswith("http"):
                iri_filter = f"FILTER (?s = <{iri}>)"
            qres = self.graph.query(
                """
                PREFIX ex: <https://example.com/>
                CONSTRUCT {
                    ?s ?p ?o .
                }
                WHERE {
                    ?s ?p ?o .
                    {{{iri_filter}}}
                }
                """.replace(
                    "{{{iri_filter}}}", iri_filter
                )
            )
            jsonld_dict = json.loads(qres.serialize(format="json-ld"))[0]
            jsonld_dicts[iri] = jsonld_dict
        return jsonld_dicts


class LocalSparqlBackend(LocalSparqlResolver, Backend):
    def store_jsonld_dicts(self, jsonld_dicts: Dict[str, Dict]) -> StoreResult:
        # delete all triples with the given iris as subject
        for iri in jsonld_dicts.keys():
            iri_filter = f"{iri}"
            # check if the iri is a full IRI or a prefix
            if iri.startswith("http"):
                iri_filter = f"<{iri}>"
            query = """
                PREFIX ex: <https://example.com/>
                DELETE WHERE {
                    {{{iri_filter}}} ?p ?o .
                }
                """.replace(
                "{{{iri_filter}}}", iri_filter
            )
            self.graph.update(query)
            # convert jsonld_dict to rdflib triples and add to graph
            g = Graph()
            g.parse(data=json.dumps(jsonld_dicts[iri]), format="json-ld")
            self.graph += g
        return StoreResult(success=True)

    def query():
        raise NotImplementedError()


class SparqlResolver(Resolver):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    endpoint: str

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._sparql = SPARQLWrapper(self.endpoint)

    def resolve_iris(self, iris: List[str]) -> Dict[str, Dict]:
        # sparql query to get a node by IRI with all its properties
        # using CONSTRUCT to get the full node
        # format the result as json-ld
        jsonld_dicts = {}

        # lookup  credential for the endpoint
        cred = get_credential(GetCredentialParam(iri=self.endpoint))
        if cred is not None:
            if isinstance(cred, UserPwdCredential):
                self._sparql.setCredentials(
                    cred.username, cred.password.get_secret_value()
                )

        for iri in iris:
            iri_filter = f"FILTER (?s = {iri})"
            # check if the iri is a full IRI or a prefix
            if iri.startswith("http"):
                iri_filter = f"FILTER (?s = <{iri}>)"
            self._sparql.setQuery(
                """
                PREFIX ex: <https://example.com/>
                CONSTRUCT {
                    ?s ?p ?o .
                }
                WHERE {
                    ?s ?p ?o .
                    {{{iri_filter}}}
                }
                """.replace(
                    "{{{iri_filter}}}", iri_filter
                )
            )
            self._sparql.setReturnFormat(JSONLD)
            result: Graph = self._sparql.query().convert()
            if len(result) == 0:
                jsonld_dicts[iri] = None
                continue
            jsonld_dict = json.loads(result.serialize(format="json-ld"))[0]
            jsonld_dicts[iri] = jsonld_dict

        return jsonld_dicts


class WikiDataSparqlResolver(Resolver):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    endpoint: str

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._sparql = SPARQLWrapper(self.endpoint)

    def resolve_iri(self, iris: List[str]) -> Dict[str, Dict]:
        # sparql query to get a node by IRI with all its properties
        # using CONSTRUCT to get the full node
        # format the result as json-ld
        jsonld_dicts = {}
        for iri in iris:
            iri_filter = f"FILTER (?s = {iri})"
            # check if the iri is a full IRI or a prefix
            if iri.startswith("http"):
                iri_filter = f"FILTER (?s = <{iri}>)"
            self._sparql.setQuery(
                """
                PREFIX ex: <https://example.com/>
                PREFIX Item: <http://www.wikidata.org/entity/>
                CONSTRUCT {
                    ?s ?p ?o .
                }
                WHERE {
                    ?s ?p ?o .
                    {{{iri_filter}}}
                }
                """.replace(
                    "{{{iri_filter}}}", iri_filter
                )
            )
            self._sparql.setReturnFormat(JSONLD)
            result: Graph = self._sparql.query().convert()
            jsonld_dict = json.loads(result.serialize(format="json-ld"))[0]
            # replace http://www.wikidata.org/prop/direct/P31 with @type
            if "http://www.wikidata.org/prop/direct/P31" in jsonld_dict:
                jsonld_dict["@type"] = jsonld_dict.pop(
                    "http://www.wikidata.org/prop/direct/P31"
                )[0]["@id"]
            jsonld_dicts[iri] = jsonld_dict

        return jsonld_dicts



# generated by datamodel-codegen:
#   filename:  example.json

from __future__ import annotations

from pydantic.v1 import Field

from oold.model.v1 import LinkedBaseModel


class Subschema1(LinkedBaseModel):
    class Config:
        schema_extra = {"custom_key": "custom_value", "title": "Subschema1"}

    subprop0: str | None = Field(None, custom_key="custom_value_0")
    subprop1: str | None = Field(None, custom_key="custom_value_1")


class Subschema2(LinkedBaseModel):
    class Config:
        schema_extra = {"custom_key": "custom_value", "title": "Subschema2"}

    subprop0: str | None = Field(None, custom_key="custom_value_0")
    subprop2: str | None = Field(None, custom_key="custom_value_2")


class Example(LinkedBaseModel):
    class Config:
        schema_extra = {"title": "Example"}

    type: str | None = ["example"]
    prop1: str | None = Field(None, custom_key="custom_value")
    prop2: Subschema1 | Subschema2 | None = Field(None, custom_key="custom_value")



# generated by datamodel-codegen:
#   filename:  example.json

from __future__ import annotations

from enum import Enum

from pydantic.v1 import Field

from oold.model.v1 import LinkedBaseModel


class Hobby(str, Enum):
    """
    Defines various hobbies as an enum.
    """

    SPORTS = "ex:sports"
    """
    Sports hobby, e.g. football, basketball, etc.
    """
    MUSIC = "ex:music"
    """
    Music hobby, e.g. playing instruments, singing, etc.
    """
    ART = "ex:art"
    """
    Art hobby, e.g. painting, drawing, etc.
    """


class Example(LinkedBaseModel):
    class Config:
        schema_extra = {"title": "Example"}

    type: str | None = ["example"]
    hobby: Hobby | None = Field(
        None,
        options={
            "enum_titles": [
                "Sports hobby, e.g. football, basketball, etc.",
                "Music hobby, e.g. playing instruments, singing, etc.",
                "Art hobby, e.g. painting, drawing, etc.",
            ]
        },
        x_enum_varnames=["SPORTS", "MUSIC", "ART"],
    )
    """
    Defines various hobbies as an enum.
    """



from typing import Optional

from oold.backend.document_store import SimpleDictDocumentStore
from oold.backend.interface import (
    Backend,
    ResolveParam,
    SetBackendParam,
    SetResolverParam,
    StoreParam,
    set_backend,
    set_resolver,
)


def _store_procedure(store: Backend, pydantic_version="v2"):
    if pydantic_version == "v1":
        # based on pydantic v1
        from oold.model.v1 import LinkedBaseModel

        class Entity(LinkedBaseModel):
            class Config:
                schema_extra = {
                    "@context": {
                        # aliases
                        "id": "@id",
                        "type": "@type",
                        # prefixes
                        "ex": "https://example.com/",
                        # literal property
                        "name": "ex:name",
                    },
                    "iri": "Entity.json",  # the IRI of the schema
                }

            type: Optional[str] = "ex:Entity.json"
            name: str

            def get_iri(self):
                return "ex:" + self.name

    else:
        from pydantic import ConfigDict

        from oold.model import LinkedBaseModel

        class Entity(LinkedBaseModel):
            model_config = ConfigDict(
                json_schema_extra={
                    "@context": {
                        # aliases
                        "id": "@id",
                        "type": "@type",
                        # prefixes
                        "schema": "https://schema.org/",
                        "ex": "https://example.com/",
                        # literal property
                        "name": "schema:name",
                    },
                    "iri": "Entity.json",  # the IRI of the schema
                }
            )

            type: Optional[str] = "ex:Entity.json"
            name: str

            def get_iri(self):
                return "ex:" + self.name

    set_resolver(SetResolverParam(iri="ex", resolver=store))
    set_backend(SetBackendParam(iri="ex", backend=store))

    e = Entity(name="TestEntity")
    store.store(StoreParam(nodes={e.get_iri(): e}))

    e2 = store.resolve(ResolveParam(iris=[e.get_iri()], model_cls=Entity)).nodes[
        e.get_iri()
    ]
    assert e2.name == "TestEntity"

    e10 = Entity(name="AnotherEntity")
    e10.store_jsonld()
    e10_retrieved = Entity["ex:AnotherEntity"]
    assert e10_retrieved.name == "AnotherEntity"


def _run(store: Backend):
    _store_procedure(store, pydantic_version="v1")
    _store_procedure(store, pydantic_version="v2")


def test_simple_dict_document_store():
    store = SimpleDictDocumentStore()
    _run(store)


def test_sqlite_document_store():
    from oold.backend.document_store import SqliteDocumentStore

    store = SqliteDocumentStore(db_path=":memory:")

    _run(store)


def test_local_sparql_store():
    from oold.backend.sparql import LocalSparqlBackend

    store = LocalSparqlBackend()

    _run(store)


if __name__ == "__main__":
    # test_simple_dict_document_store()
    # test_sqlite_document_store()
    test_local_sparql_store()



from typing import List, Optional

from pydantic import ConfigDict, Field

from oold.backend.interface import SetResolverParam, set_resolver
from oold.backend.sparql import WikiDataSparqlResolver

# based on pydantic v2
from oold.model import LinkedBaseModel  # noqa


class MultiLanguageString(LinkedBaseModel):
    text: str
    lang: str


class WikiDataEntity(LinkedBaseModel):
    model_config = ConfigDict(
        json_schema_extra={
            "@context": {
                # aliases
                "id": "@id",
                # prefixes
                "p": "http://www.wikidata.org/prop/",
                "wdt": "http://www.wikidata.org/prop/direct/",
                "Item": "http://www.wikidata.org/entity/",
                "type": "wdt:P31",
                "name": {
                    "@id": "wdt:P373",
                    "@type": "http://www.w3.org/2001/XMLSchema#string",
                },
                # "label": {
                #     "@id": "http://www.w3.org/2000/01/rdf-schema#label",
                #     "@container": "@set",
                #     "@context": {
                #         "text": "@value",
                #         "lang": "@language",
                #     }
                # },
            },
            "iri": "Entity.json",  # the IRI of the schema
        }
    )
    id: str
    type: Optional[str]
    # label: Optional[List[MultiLanguageString]] = None
    name: Optional[str] = None

    @classmethod
    def get_class_iri(cls):
        # return default value of field 'type' if not set
        if (
            cls.model_fields.get("type")
            and cls.model_fields["type"].default is not None
        ):
            return cls.model_fields["type"].default

    def get_iri(self):
        return "ex:" + self.name


class Person(WikiDataEntity):  # noqa
    model_config = ConfigDict(
        json_schema_extra={
            "@context": [
                "Entity.json",  # import the context of the parent class
                {
                    # object property definition
                    "father": {
                        "@id": "wdt:P22",
                        "@type": "@id",
                    },
                    "knows": {
                        "@id": "schema:knows",
                        "@type": "@id",
                        "@container": "@set",
                    },
                },
            ],
            "iri": "Q5",
        }
    )
    type: Optional[str] = "wd:Q5"  # Q5 is the Wikidata item for human
    father: Optional["Person"] = Field(
        None,
        json_schema_extra={"range": "Person.json"},
    )
    knows: Optional[List["Person"]] = Field(
        None,
        # object property pointing to another Person
        json_schema_extra={"range": "Person.json"},
    )


# create a resolver to resolve IRIs to objects


r = WikiDataSparqlResolver(endpoint="https://query.wikidata.org/sparql")
set_resolver(SetResolverParam(iri="Item", resolver=r))

# Example usage:
p = Person["Item:Q80"]  # Douglas Adams
print(p.model_dump_json(indent=2))
print(p)
print(p.father)
print(p.father.father)



# generated by datamodel-codegen:
#   filename:  Person.json

from __future__ import annotations

from pydantic.v1 import Field, constr

from oold.model.v1 import LinkedBaseModel


class Thing(LinkedBaseModel):
    class Config:
        schema_extra = {"title": "Thing"}

    type: str | None = Field("playground:Thing", options={"hidden": True})
    name: constr(min_length=1) | None = "A Thing"
    """
    The things name
    """


class Person(Thing):
    class Config:
        schema_extra = {"title": "Person"}

    type: str | None = Field(
        "playground:Person",
        field_comment="Already defined in playground:Thing -> we override just the default",
    )
    name: constr(min_length=1) | None = "John Doe"
    """
    First and Last name
    """
    age: int | None = None



import json
import sqlite3
from pathlib import Path
from typing import Dict, List, Optional, Union

from oold.backend.interface import Backend, StoreResult


class SimpleDictDocumentStore(Backend):
    _store: Optional[Dict[str, dict]] = None

    def __init__(self):
        self._store = {}

    def resolve_iris(self, iris: List[str]) -> Dict[str, Dict]:
        jsonld_dicts = {}
        for iri in iris:
            jsonld_dicts[iri] = self._store.get(iri, None)
        return jsonld_dicts

    def store_jsonld_dicts(self, jsonld_dicts: Dict[str, Dict]) -> StoreResult:
        for iri, jsonld_dict in jsonld_dicts.items():
            self._store[iri] = jsonld_dict
        return StoreResult(success=True)

    def query():
        pass


class SqliteDocumentStore(Backend):
    db_path: Union[Path, str]
    persist_connection: bool = False
    _conn: Optional[sqlite3.Connection] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if self.db_path == ":memory:":
            self.persist_connection = True
            self._conn = sqlite3.connect(self.db_path)

        # create table 'entities' if not exists
        conn = self._conn if self.persist_connection else sqlite3.connect(self.db_path)
        c = conn.cursor()
        c.execute(
            """
            CREATE TABLE IF NOT EXISTS entities (
                id TEXT PRIMARY KEY,
                data JSONB
            )
            """
        )
        conn.commit()
        if not self.persist_connection:
            conn.close()

    def close(self):
        """Close the persistent connection if it exists."""
        if self._conn is not None:
            self._conn.close()
            self._conn = None

    def resolve_iris(self, iris: List[str]) -> Dict[str, Dict]:
        jsonld_dicts = {}
        conn = self._conn if self.persist_connection else sqlite3.connect(self.db_path)
        c = conn.cursor()
        c.execute(
            """
            SELECT id, data FROM entities WHERE id IN ({})
            """.format(
                ",".join("?" for _ in iris)
            ),
            iris,
        )
        rows = c.fetchall()
        for iri, data in rows:
            jsonld_dicts[iri] = json.loads(data)
        if not self.persist_connection:
            conn.close()
        return jsonld_dicts

    def store_jsonld_dicts(self, jsonld_dicts: Dict[str, Dict]) -> StoreResult:
        conn = self._conn if self.persist_connection else sqlite3.connect(self.db_path)
        c = conn.cursor()
        c.executemany(
            """
            INSERT OR REPLACE INTO entities (id, data) VALUES (?, ?)
            """,
            [
                (iri, json.dumps(jsonld_dict))
                for iri, jsonld_dict in jsonld_dicts.items()
            ],
        )
        conn.commit()
        if not self.persist_connection:
            conn.close()
        return StoreResult(success=True)

    def query():
        raise NotImplementedError()



# generated by datamodel-codegen:
#   filename:  Entity.json

from __future__ import annotations

from pydantic import ConfigDict, Field

from oold.model import LinkedBaseModel


class SimpleSubSchema(LinkedBaseModel):
    model_config = ConfigDict(
        json_schema_extra={"title": "SimpleSubSchema"},
    )
    some_property: str | None = None


class NestedSubSchema(LinkedBaseModel):
    """
    An example subschema
    """

    model_config = ConfigDict(
        json_schema_extra={
            "title": "NestedSubSchema",
            "description": "An example subschema",
        },
    )
    subschema: SimpleSubSchema | None = Field(None, title="SimpleSubSchema")


class Entity(LinkedBaseModel):
    model_config = ConfigDict(
        json_schema_extra={"title": "Entity"},
    )
    nested: NestedSubSchema | None = Field(
        None,
        json_schema_extra={"x-custom-annotation": "custom value"},
        title="EntityNestedSubSchema",
    )
    """
    An example using a subschema
    """



from abc import abstractmethod
from typing import Dict, List, Optional, Type, Union

from pydantic import BaseModel

from oold.static import GenericLinkedBaseModel


class SetResolverParam(BaseModel):
    iri: str
    resolver: "Resolver"


class GetResolverParam(BaseModel):
    iri: str


class GetResolverResult(BaseModel):
    resolver: "Resolver"


class ResolveParam(BaseModel):
    iris: List[str]
    model_cls: Optional[Type[GenericLinkedBaseModel]] = None


class ResolveResult(BaseModel):
    model_config = {
        "arbitrary_types_allowed": True,
    }
    nodes: Dict[str, Union[None, GenericLinkedBaseModel]]


class Resolver(BaseModel):
    model_cls: Optional[Type[GenericLinkedBaseModel]] = None

    @abstractmethod
    def resolve_iris(self, iris: List[str]) -> Dict[str, Dict]:
        pass

    def resolve(self, request: ResolveParam):
        # print("RESOLVE", request)

        model_cls = request.model_cls
        if model_cls is None:
            model_cls = self.model_cls
        if model_cls is None:
            raise ValueError("No model_cls provided in request or resolver")

        jsonld_dicts = self.resolve_iris(request.iris)
        nodes = {}
        for iri, jsonld_dict in jsonld_dicts.items():
            if jsonld_dict is None:
                nodes[iri] = None
            else:
                node = model_cls.from_jsonld(jsonld_dict)
                nodes[iri] = node

        return ResolveResult(nodes=nodes)


global _resolvers
_resolvers = {}


def set_resolver(param: SetResolverParam) -> None:
    _resolvers[param.iri] = param.resolver


def get_resolver(param: GetResolverParam) -> GetResolverResult:
    # ToDo: Handle prefixes (ex:) as well as full IRIs (http://example.com/)
    # ToDo: Handle list of IRIs with mixed domains
    iri = param.iri.split(":")[0]
    if iri not in _resolvers:
        raise ValueError(f"No resolvers found for {iri}")
    return GetResolverResult(resolver=_resolvers[iri])


class SetBackendParam(BaseModel):
    iri: str
    backend: "Backend"


class GetBackendParam(BaseModel):
    iri: str


class GetBackendResult(BaseModel):
    backend: "Backend"


class StoreParam(BaseModel):
    model_config = {
        "arbitrary_types_allowed": True,
    }
    nodes: Dict[str, Union[None, GenericLinkedBaseModel]]


class StoreResult(BaseModel):
    success: bool


class Query(BaseModel):
    pass


class Backend(Resolver):
    def store(self, param: StoreParam) -> StoreResult:
        jsonld_dicts = {}
        for iri, node in param.nodes.items():
            if node is None:
                jsonld_dicts[iri] = None
            else:
                jsonld_dicts[iri] = node.to_jsonld()
        return self.store_jsonld_dicts(jsonld_dicts)

    @abstractmethod
    def store_jsonld_dicts(self, jsonld_dicts: Dict[str, Dict]) -> StoreResult:
        pass

    @abstractmethod
    def query(self, query: Query) -> ResolveResult:
        """Query the backend and return a ResolveResult."""
        pass


global _backends
_backends = {}


def set_backend(param: SetBackendParam) -> None:
    _backends[param.iri] = param.backend


def get_backend(param: GetBackendParam) -> GetBackendResult:
    iri = param.iri.split(":")[0]
    if iri not in _backends:
        raise ValueError(f"No backends found for {iri}")
    return GetBackendResult(backend=_backends[iri])



from oold.backend.auth import SetCredentialParam, UserPwdCredential, set_credential
from oold.backend.sparql import SparqlResolver

# ToDo: allow other ways, e.g. environment variables, keyring, ...
set_credential(
    SetCredentialParam(
        credential=UserPwdCredential(
            iri="https://blazegraph.kiprobatt.de",
            username="user",
            password="*********",
        )
    )
)

r = SparqlResolver(
    endpoint="https://blazegraph.kiprobatt.de/blazegraph/namespace/kb/sparql"
)

# this will lookup the credential and use it for authentication
# raise an error if the credential is missing or invalid
r.resolve_iris(["http://example.com/Entity1"])



import getpass
from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel, SecretStr

global _credentials
_credentials = {}


class BaseCredential(BaseModel):
    """Abstract base class for credentials"""

    iri: Union[str, List[str]]
    """the IRI(s) the credential is valid for"""


class UserPwdCredential(BaseCredential):
    """a username - password credential"""

    username: str
    """the user identifier"""
    password: SecretStr
    """the users password"""


class CredentialFallback(str, Enum):
    """Modes of handling missing credentials

    Attributes
    ----------
    none:
        throw error
    ask:
        use getpass to ask for credentials
    """

    ask = "ask"  # use getpass to ask for credentials
    none = "none"  # throw error


class GetCredentialParam(BaseModel):
    """Reads credentials from a yaml file"""

    iri: str
    """internationalized resource identifier / address of the service, may contain
    protocol, domain, port and path matches by "contains" returning the shortest
    match"""
    fallback: Optional[CredentialFallback] = CredentialFallback.none
    """The fallback strategy if no credential was found for the given origin"""


class SetCredentialParam(BaseModel):
    credential: BaseCredential


def set_credential(param: SetCredentialParam) -> None:
    global _credentials
    iris = (
        param.credential.iri
        if isinstance(param.credential.iri, list)
        else [param.credential.iri]
    )
    for iri in iris:
        _credentials[iri] = param.credential


def get_credential(param: GetCredentialParam) -> BaseCredential:
    match_iri = ""
    cred = None
    iris = _credentials.keys()
    for iri in iris:
        if iri in param.iri:
            if match_iri == "" or len(match_iri) > len(
                iri
            ):  # use the less specific match
                match_iri = iri

    if match_iri != "":
        cred = _credentials.get(match_iri, None)

    if cred is None:
        if param.fallback is CredentialFallback.ask:
            print(
                f"No credentials for {param.iri} found. "
                f"Please use the prompt to login"
            )
            username = input("Enter username: ")
            password = getpass.getpass("Enter password: ")
            cred = UserPwdCredential(
                username=username, password=password, iri=param.iri
            )
            set_credential(SetCredentialParam(credential=cred))
    return cred



# generated by datamodel-codegen:
#   filename:  Person.json

from __future__ import annotations

from pydantic import ConfigDict, Field, constr

from oold.model import LinkedBaseModel


class Thing(LinkedBaseModel):
    model_config = ConfigDict(
        json_schema_extra={"title": "Thing"},
    )
    type: str | None = Field(
        "playground:Thing", json_schema_extra={"options": {"hidden": True}}
    )
    name: constr(min_length=1) | None = "A Thing"
    """
    The things name
    """


class Person(Thing):
    model_config = ConfigDict(
        json_schema_extra={"title": "Person"},
    )
    type: str | None = Field(
        "playground:Person",
        json_schema_extra={
            "$comment": "Already defined in playground:Thing -> we override just the default"
        },
    )
    name: constr(min_length=1) | None = "John Doe"
    """
    First and Last name
    """
    age: int | None = None
